import platform
import gc
import pytest

is_pypy = platform.python_implementation() == 'PyPy'

def collect():
    if is_pypy:
        for i in range(3):
            gc.collect()
    else:
        gc.collect()

skip_on_pypy = pytest.mark.skipif(
    is_pypy, reason="This test currently fails/crashes PyPy")



import torch
import cuda_mock
from cuda_mock import *

script_dir = os.path.dirname(os.path.abspath(__file__))

lib = DynamicObj(f'{script_dir}/test_hook_strlen.cnn').compile().appen_compile_opts('-lpthread').get_lib()

cuda_mock.internal_install_hook("libc.so", "cuda_mock_impl.cpython-38-x86_64-linux-gnu.so", "strlen", str(lib), "strlen")



import sys
import test_classes_ext as t
import pytest
from common import skip_on_pypy, collect


@pytest.fixture
def clean():
    collect()
    t.reset()


def assert_stats(**kwargs):
    collect()
    for k, v in t.stats().items():
        fail = False
        if k in kwargs:
            if v != kwargs[k]:
                fail = True
        elif v != 0:
            fail = True
        if fail:
            raise Exception(f'Mismatch for key {k}: {t.stats()}')

def test01_signature():
    assert t.Struct.__init__.__doc__ == (
        "__init__(self) -> None\n"
        "__init__(self, arg: int, /) -> None"
    )

    assert t.Struct.value.__doc__ == "value(self) -> int"
    assert t.Struct.create_move.__doc__ == "create_move() -> test_classes_ext.Struct"
    assert t.Struct.set_value.__doc__ == "set_value(self, value: int) -> None"
    assert t.Struct().set_value.__doc__ == "set_value(self, value: int) -> None"
    assert t.Struct.__doc__ == 'Some documentation'
    assert t.Struct.static_test.__doc__ == (
        "static_test(arg: int, /) -> int\n"
        "static_test(arg: float, /) -> int")


def test02_static_overload():
    assert t.Struct.static_test(1) == 1
    assert t.Struct.static_test(1.0) == 2


def test03_instantiate(clean):
    s1 : t.Struct = t.Struct()
    assert s1.value() == 5
    s2 = t.Struct(10)
    assert s2.value() == 10
    del s1
    del s2
    assert_stats(
        default_constructed=1,
        value_constructed=1,
        destructed=2
    )

def test04_double_init():
    s = t.Struct()
    with pytest.warns(RuntimeWarning, match='nanobind: attempted to initialize an already-initialized instance of type'):
        with pytest.raises(TypeError):
            s.__init__(3)


def test05_rv_policy(clean):
    s = t.Struct()
    assert s.self() is s
    assert s.none() is None
    del s
    assert_stats(
        default_constructed=1,
        destructed=1
    )

    # ------

    t.reset()
    assert t.Struct.create_take().value() == 10
    assert_stats(
        value_constructed=1,
        destructed=1
    )

    # ------

    t.reset()
    assert t.Struct.create_move().value() == 11
    assert_stats(
        value_constructed=1,
        move_constructed=1,
        destructed=2
    )

    # ------

    t.reset()
    assert t.Struct.create_reference().value() == 12
    assert_stats()

    # ------

    t.reset()
    assert t.Struct.create_copy().value() == 12
    assert_stats(
        copy_constructed=1,
        destructed=1
    )

def test06_reference_internal(clean):
    s = t.PairStruct()
    s1 = s.s1
    s2 = s.s2
    del s
    assert_stats(default_constructed=2)
    assert s2.value() == 5
    del s2

    assert_stats(default_constructed=2)

    assert s1.value() == 5
    del s1
    assert_stats(
        default_constructed=2,
        destructed=2
    )

    # ----

    s = t.PairStruct()
    s1 = s.s1
    del s1
    del s


def test07_big():
    x = [t.Big() for i in range(1024)]
    x2 = [t.BigAligned() for i in range(1024)]


def test08_inheritance():
    dog = t.Dog('woof')
    cat = t.Cat('meow')
    assert dog.name() == 'Dog'
    assert cat.name() == 'Cat'
    assert dog.what() == 'woof'
    assert cat.what() == 'meow'
    assert isinstance(dog, t.Animal) and isinstance(dog, t.Dog)
    assert isinstance(cat, t.Animal) and isinstance(cat, t.Cat)
    assert t.go(dog) == 'Dog says woof'
    assert t.go(cat) == 'Cat says meow'

    assert t.animal_passthrough(dog) is dog
    assert t.animal_passthrough(cat) is cat
    assert t.dog_passthrough(dog) is dog

    with pytest.raises(TypeError):
        t.dog_passthrough(cat)

def test09_method_vectorcall():
    out = []

    def f(a, b, c, d, e):
        out.append((a, b, c, d, e))

    class MyClass:
        def f(self, a, b, c, d, e):
            self.out = ((a, b, c, d, e))

    t.call_function(f)

    i = MyClass()
    t.call_method(i)
    assert out == [(1, 2, "hello", True, 4)]
    assert i.out == (1, 2, "hello", True, 4)


def test10_trampoline(clean):
    for _ in range(10):
        class Dachshund(t.Animal):
            def __init__(self):
                super().__init__()
            def name(self):
                return "Dachshund"
            def what(self):
                return "yap"

        d = Dachshund()
        for _ in range(10):
            assert t.go(d) == 'Dachshund says yap'

        assert t.animal_passthrough(d) is d

    a = 0
    class GenericAnimal(t.Animal):
        def what(self):
            return "goo"

        def void_ret(self):
            nonlocal a
            a += 1

        def name(self):
            return 'Generic' + super().name()

    ga = GenericAnimal()
    assert t.go(ga) == 'GenericAnimal says goo'
    assert t.void_ret(ga) is None
    assert a == 1
    assert t.animal_passthrough(ga) is ga

    del ga
    del d

    assert_stats(
        default_constructed=11,
        destructed=11
    )

    class GenericDog(t.Dog):
        pass

    d = GenericDog("GenericDog")
    assert t.dog_passthrough(d) is d
    assert t.animal_passthrough(d) is d

def test11_trampoline_failures():
    class Incomplete(t.Animal):
        def __init__(self):
            super().__init__()

        def void_ret(self):
            raise TypeError("propagating an exception")

    d = Incomplete()
    with pytest.raises(RuntimeError) as excinfo:
        t.go(d)
    assert ('test_classes.Incomplete::what()\'): tried to call a pure virtual function!' in str(excinfo.value))

    with pytest.raises(TypeError) as excinfo:
        t.void_ret(d)
    assert 'propagating an exception' in str(excinfo.value)

    class Incomplete2(t.Animal):
        def __init__(self):
            pass # Missing call to super().__init__()
        def name(self):
            return "a"
        def what(self):
            return "b"

    with pytest.warns(RuntimeWarning, match='nanobind: attempted to access an uninitialized instance of type'):
        with pytest.raises(TypeError) as excinfo:
            t.go(Incomplete2())
    assert 'incompatible function arguments' in str(excinfo.value)


def test12_large_pointers():
    import struct
    for i in range(1, 10):
        c = t.i2p(i)
        assert isinstance(c, t.Foo)
        assert t.p2i(c) == i

    large = (1 << (struct.calcsize("P")*8))-1
    for i in range(large - 10, large):
        c = t.i2p(i)
        assert isinstance(c, t.Foo)
        assert t.p2i(c) == i


def test13_implicitly_convertible():
    assert t.get_d.__doc__ == "get_d(arg: test_classes_ext.D, /) -> int"
    a = t.A(1)
    b = t.B(2)
    b2 = t.B2(3)
    c = t.C(4)
    d = 5
    with pytest.raises(TypeError) as excinfo:
        t.get_d(c)
    assert str(excinfo.value) == (
        "get_d(): incompatible function arguments. The following argument types are supported:\n"
        "    1. get_d(arg: test_classes_ext.D, /) -> int\n"
        "\n"
        "Invoked with types: test_classes_ext.C")
    assert t.get_d(a) == 11
    assert t.get_d(b) == 102
    assert t.get_d(b2) == 103
    assert t.get_d(d) == 10005


def test14_operators():
    a = t.Int(1)
    b = t.Int(2)
    assert repr(a + b) == "3"
    with pytest.raises(TypeError) as excinfo:
        assert repr(a - b) == "3"
    assert "unsupported operand type" in str(excinfo.value)
    assert repr(a - 2) == "-1"
    a += b
    assert repr(a) == "3"
    assert repr(b) == "2"

    assert a.__add__("test") is NotImplemented


def test15_keep_alive_nbtype(clean):
    t.reset()
    s = t.Struct()
    a = t.Dog('Rufus')
    assert t.keep_alive_arg(s, a) is a
    assert t.keep_alive_arg(s, a) is a
    del s
    assert_stats(
        default_constructed=1
    )
    del a
    assert_stats(
        default_constructed=1,
        destructed=1
    )

    t.reset()
    s = t.Struct()
    a = t.Dog('Rufus')
    assert t.keep_alive_ret(a, s) is s
    assert t.keep_alive_ret(a, s) is s
    del a
    assert_stats(
        default_constructed=1
    )
    del s
    assert_stats(
        default_constructed=1,
        destructed=1
    )


def test16_keep_alive_custom(clean):
    constructed = 0
    destructed = 0

    class Struct():
        def __init__(self):
            nonlocal constructed
            constructed += 1

        def __del__(self):
            nonlocal destructed
            destructed += 1

    class Struct2():
        def __init__(self):
            pass

    s = Struct()
    a = Struct2()
    assert t.keep_alive_arg(s, a) is a
    del s
    collect()
    assert constructed == 1 and destructed == 0
    del a
    collect()
    assert constructed == 1 and destructed == 1

    s = Struct()
    a = Struct2()
    assert t.keep_alive_ret(a, s) is s
    del a
    collect()
    assert constructed == 2 and destructed == 1
    del s
    collect()
    assert constructed == 2 and destructed == 2

    with pytest.raises(RuntimeError) as excinfo:
        s = Struct()
        x = 5
        t.keep_alive_ret(x, s)

    assert "nanobind::detail::keep_alive(): could not create a weak reference!" in str(excinfo.value)

def f():
    pass

class MyClass:
    def f(self):
        pass

    class NestedClass:
        def f(self):
            pass

def test17_name_qualname_module():
    # First, check what CPython does
    assert f.__module__ == 'test_classes'
    assert f.__name__ == 'f'
    assert f.__qualname__ == 'f'
    assert MyClass.__name__ == 'MyClass'
    assert MyClass.__qualname__ == 'MyClass'
    assert MyClass.__module__ == 'test_classes'
    assert MyClass.f.__name__ == 'f'
    assert MyClass.f.__qualname__ == 'MyClass.f'
    assert MyClass.f.__module__ == 'test_classes'
    assert MyClass().f.__name__ == 'f'
    assert MyClass().f.__qualname__ == 'MyClass.f'
    assert MyClass().f.__module__ == 'test_classes'
    assert MyClass.NestedClass.__name__ == 'NestedClass'
    assert MyClass.NestedClass.__qualname__ == 'MyClass.NestedClass'
    assert MyClass.NestedClass.__module__ == 'test_classes'
    assert MyClass.NestedClass.f.__name__ == 'f'
    assert MyClass.NestedClass.f.__qualname__ == 'MyClass.NestedClass.f'
    assert MyClass.NestedClass.f.__module__ == 'test_classes'

    # Now, check the extension module
    assert t.f.__module__ == 'test_classes_ext'
    assert t.f.__name__ == 'f'
    assert t.f.__qualname__ == 'f'
    assert type(t.f).__module__ == 'nanobind'
    assert type(t.f).__name__ == 'nb_func'
    assert type(t.f).__qualname__ == 'nb_func'
    assert t.MyClass.__name__ == 'MyClass'
    assert t.MyClass.__qualname__ == 'MyClass'
    assert t.MyClass.__module__ == 'test_classes_ext'
    assert t.MyClass.f.__name__ == 'f'
    assert t.MyClass.f.__qualname__ == 'MyClass.f'
    assert t.MyClass.f.__module__ == 'test_classes_ext'
    assert t.MyClass().f.__name__ == 'f'
    assert t.MyClass().f.__qualname__ == 'MyClass.f'
    assert t.MyClass().f.__module__ == 'test_classes_ext'
    assert type(t.MyClass.f).__module__ == 'nanobind'
    assert type(t.MyClass.f).__name__ == 'nb_method'
    assert type(t.MyClass.f).__qualname__ == 'nb_method'
    assert t.MyClass.NestedClass.__name__ == 'NestedClass'
    assert t.MyClass.NestedClass.__qualname__ == 'MyClass.NestedClass'
    assert t.MyClass.NestedClass.__module__ == 'test_classes_ext'
    assert t.MyClass.NestedClass.f.__name__ == 'f'
    assert t.MyClass.NestedClass.f.__qualname__ == 'MyClass.NestedClass.f'
    assert t.MyClass.NestedClass.f.__module__ == 'test_classes_ext'


def test18_static_properties():
    assert t.StaticProperties.value == 23
    t.StaticProperties.value += 1
    assert t.StaticProperties.value == 24
    assert t.StaticProperties.get() == 24
    assert t.StaticProperties2.get() == 24
    t.StaticProperties2.value = 50
    assert t.StaticProperties2.get() == 50
    assert t.StaticProperties.get() == 50


@skip_on_pypy
def test19_static_properties_doc():
    import pydoc
    assert "Static property docstring" in pydoc.render_doc(t.StaticProperties2)


def test20_supplement():
    c = t.ClassWithSupplement()
    assert t.check_supplement(c)
    assert not t.check_supplement(t.Struct())


def test21_type_callback():
    o = t.ClassWithLen()
    assert len(o) == 123


def test22_low_level(clean):
    s1, s2, s3, s4 = t.test_lowlevel()
    assert s1.value() == 123 and s2.value() == 0 and s4.value() == 123
    assert s3.s1.value() == 123 and s3.s2.value() == 456
    assert s3.s1 is s4
    del s1
    del s2
    del s3
    import gc
    gc.collect()
    gc.collect()
    assert s4.value() == 123
    del s4
    assert_stats(
        value_constructed=3,
        copy_constructed=1,
        move_constructed=1,
        destructed=5
    )


def test23_handle_t(clean):
    assert t.test_handle_t.__doc__ == 'test_handle_t(arg: test_classes_ext.Struct, /) -> object'
    s = t.test_handle_t(t.Struct(5))
    assert s.value() == 5
    del s

    with pytest.raises(TypeError) as excinfo:
        t.test_handle_t("test")
    assert "incompatible function argument" in str(excinfo.value)
    assert_stats(
        value_constructed=1,
        destructed=1
    )


def test24_type_object_t(clean):
    if sys.version_info < (3, 9):
        assert t.test_type_object_t.__doc__ == 'test_type_object_t(arg: Type[test_classes_ext.Struct], /) -> object'
    else:
        assert t.test_type_object_t.__doc__ == 'test_type_object_t(arg: type[test_classes_ext.Struct], /) -> object'

    assert t.test_type_object_t(t.Struct) is t.Struct

    with pytest.raises(TypeError):
        t.test_type_object_t(t.Struct())

    with pytest.raises(TypeError):
        t.test_type_object_t(int)


def test25_none_arg():
    with pytest.raises(TypeError):
        t.none_0(None)
    with pytest.raises(TypeError):
        t.none_1(None)
    with pytest.raises(TypeError):
        t.none_2(arg=None)
    assert t.none_3(None) is True
    assert t.none_4(arg=None) is True
    assert t.none_0.__doc__ == 'none_0(arg: test_classes_ext.Struct, /) -> bool'
    assert t.none_1.__doc__ == 'none_1(arg: test_classes_ext.Struct) -> bool'
    assert t.none_2.__doc__ == 'none_2(arg: test_classes_ext.Struct) -> bool'
    assert t.none_3.__doc__ == 'none_3(arg: Optional[test_classes_ext.Struct]) -> bool'
    assert t.none_4.__doc__ == 'none_4(arg: Optional[test_classes_ext.Struct]) -> bool'


def test26_is_final():
    with pytest.raises(TypeError) as excinfo:
        class MyType(t.FinalType):
            pass
    assert "The type 'test_classes_ext.FinalType' prohibits subclassing!" in str(excinfo.value)


def test27_dynamic_attr(clean):
    l = [None] * 100
    for i in range(100):
        l[i] = t.StructWithAttr(i)

    # Create a big reference cycle..
    for i in range(100):
        l[i].prev = l[i - 1]
        l[i].next = l[i + 1 if i < 99 else 0]
        l[i].t = t.StructWithAttr
        l[i].self = l[i]

    for i in range(100):
        assert l[i].value() == i
        assert l[i].self.value() == i
        assert l[i].prev.value() == (i-1 if i > 0 else 99)
        assert l[i].next.value() == (i+1 if i < 99 else 0)

    del l

    assert_stats(
        value_constructed=100,
        destructed=100
    )


def test28_copy_rvp():
    a = t.Struct.create_reference()
    b = t.Struct.create_copy()
    assert a is not b


def test29_pydoc():
    import pydoc
    assert "Some documentation" in pydoc.render_doc(t)


def test30_property_assignment_instance():
    s = t.PairStruct()
    s1 = t.Struct(123)
    s2 = t.Struct(456)
    s.s1 = s1
    s.s2 = s2
    assert s2 is not s.s2 and s1 is not s.s1
    assert s.s1.value() == 123
    assert s.s2.value() == 456
    assert s1.value() == 123
    assert s2.value() == 456


# cpyext reference cycles are not supported, see https://foss.heptapod.net/pypy/pypy/-/issues/3849
@skip_on_pypy()
def test31_cycle():
    a = t.Wrapper()
    a.value = a
    del a
    collect()


def test32_type_checks():
    v1 = 5
    v2 = t.Struct()

    assert t.is_int_1(v1) and not t.is_int_1(v2)
    assert t.is_int_2(v1) and not t.is_int_2(v2)
    assert not t.is_struct(v1) and t.is_struct(v2)


def test33_polymorphic_downcast():
    assert isinstance(t.factory(), t.Base)
    assert isinstance(t.factory_2(), t.Base)
    assert isinstance(t.polymorphic_factory(), t.PolymorphicSubclass)
    assert isinstance(t.polymorphic_factory_2(), t.PolymorphicBase)


def test34_trampoline_optimization():
    class Rufus(t.Dog):
        def __init__(self):
            super().__init__("woof")

        def name(self):
            return "Rufus"

    for i in range(2):
        d1 = t.Dog("woof")
        d2 = Rufus()

        if i == 0:
            assert t.go(d1) == 'Dog says woof'
            assert t.go(d2) == 'Rufus says woof'

        old = t.Dog.name
        try:
            t.Dog.name = lambda self: "Max"

            assert t.go(d1) == 'Dog says woof'
            assert t.go(d2) == 'Rufus says woof'
        finally:
            t.Dog.name = old


def test35_method_introspection():
    obj = t.Struct(5)
    m = obj.value
    assert m() == m.__call__() == 5
    assert hash(m) == m.__hash__()
    assert repr(m) == m.__repr__()
    assert "bound_method" in repr(m)
    assert m.__self__ is obj
    assert m.__func__ is t.Struct.value
    # attributes not defined by nb_bound_method are forwarded to nb_method:
    assert m.__name__ == "value"
    assert m.__qualname__ == "Struct.value"
    assert m.__module__ == t.__name__
    assert m.__doc__ == t.Struct.value.__doc__ == "value(self) -> int"


def test38_pickle(clean):
    import pickle

    s = t.Struct(123)
    s2 = pickle.dumps(s, protocol=pickle.HIGHEST_PROTOCOL)
    s3 = pickle.loads(s2)
    assert s.value() == s3.value()
    del s, s3

    assert_stats(
        value_constructed=1,
        pickled=1,
        unpickled=1,
        destructed=2
    )

def test39_try_cast(clean):
    s = t.Struct(123)

    assert_stats(value_constructed=1)
    t.reset()

    rv, s2 = t.try_cast_1(s)
    assert rv is True and s2 is not s and s.value() == 123 and s2.value() == 123
    del s2
    assert_stats(default_constructed=1, move_constructed=2, copy_assigned=1, destructed=3)
    t.reset()

    rv, s2 = t.try_cast_2(s)
    assert rv is True and s2 is not s and s.value() == 123 and s2.value() == 123
    del s2
    assert_stats(default_constructed=1, move_constructed=2, copy_assigned=1, destructed=3)
    t.reset()

    rv, s2 = t.try_cast_3(s)
    assert rv is True and s2 is s and s.value() == 123
    del s2
    assert_stats()
    t.reset()

    rv, s2 = t.try_cast_2(1)
    assert rv is False
    del s2
    assert_stats(default_constructed=1, move_constructed=2, destructed=3)
    t.reset()

    rv, s2 = t.try_cast_3(1)
    assert rv is False and s2 is None
    del s2
    assert_stats()
    t.reset()

    rv, s2 = t.try_cast_4(s)
    assert rv is False and s2 == 0
    rv, s2 = t.try_cast_4(123)
    assert rv is True and s2 is 123
    del s, s2

    assert_stats(destructed=1)

def test40_slots():
    if not hasattr(t, "test_slots"):
        pytest.skip()
    assert t.test_slots() == (True, True, True)



import test_stl_ext as t
import pytest
import sys
from common import collect, skip_on_pypy


@pytest.fixture
def clean():
    collect()
    t.reset()


def assert_stats(**kwargs):
    collect()
    for k, v in t.stats().items():
        fail = False
        if k in kwargs:
            if v != kwargs[k]:
                fail = True
        elif v != 0:
            fail = True
        if fail:
            raise Exception(f'Mismatch for key {k}: {t.stats()}')


# ------------------------------------------------------------------
# The following aren't strictly STL tests, but they are helpful in
# ensuring that move constructors/copy constructors of bound C++ types
# are properly triggered, which the STL type casters depend on.
# ------------------------------------------------------------------

def test01_movable_return(clean):
    assert t.return_movable().value == 5
    assert_stats(
        default_constructed=1,
        move_constructed=1,
        destructed=2)


def test02_movable_return_ptr(clean):
    assert t.return_movable_ptr().value == 5
    assert_stats(
        default_constructed=1,
        destructed=1)


def test03_movable_in_value(clean):
    s = t.Movable()
    t.movable_in_value(s)
    assert s.value == 5
    del s
    assert_stats(
        default_constructed=1,
        copy_constructed=1,
        destructed=2)


def test04_movable_in_lvalue_ref(clean):
    s = t.Movable()
    t.movable_in_lvalue_ref(s)
    assert s.value == 5
    del s
    assert_stats(
        default_constructed=1,
        destructed=1)


def test05_movable_in_ptr(clean):
    s = t.Movable()
    t.movable_in_ptr(s)
    assert s.value == 5
    del s
    assert_stats(
        default_constructed=1,
        destructed=1)


def test06_movable_in_rvalue_ref(clean):
    s = t.Movable()
    t.movable_in_rvalue_ref(s)
    assert s.value == 0
    del s
    assert_stats(
        default_constructed=1,
        move_constructed=1,
        destructed=2)


def test07_copyable_return(clean):
    assert t.return_copyable().value == 5
    assert_stats(
        default_constructed=1,
        copy_constructed=1,
        destructed=2)


def test08_copyable_return_ptr(clean):
    assert t.return_copyable_ptr().value == 5
    assert_stats(
        default_constructed=1,
        destructed=1)


def test09_copyable_in_value(clean):
    s = t.Copyable()
    t.copyable_in_value(s)
    assert s.value == 5
    del s
    assert_stats(
        default_constructed=1,
        copy_constructed=1,
        destructed=2)


def test10_copyable_in_lvalue_ref(clean):
    s = t.Copyable()
    t.copyable_in_lvalue_ref(s)
    assert s.value == 5
    del s
    assert_stats(
        default_constructed=1,
        destructed=1)


def test11_copyable_in_ptr(clean):
    s = t.Copyable()
    t.copyable_in_ptr(s)
    assert s.value == 5
    del s
    assert_stats(
        default_constructed=1,
        destructed=1)


def test12_copyable_in_rvalue_ref(clean):
    s = t.Copyable()
    t.copyable_in_rvalue_ref(s)
    assert s.value == 5
    del s
    assert_stats(
        default_constructed=1,
        copy_constructed=1,
        destructed=2)

# ------------------------------------------------------------------

def test13_tuple_movable_return(clean):
    assert t.tuple_return_movable()[0].value == 5
    assert_stats(
        default_constructed=1,
        move_constructed=2,
        destructed=3)


def test14_tuple_movable_return_ptr(clean):
    assert t.return_movable_ptr().value == 5
    assert_stats(
        default_constructed=1,
        destructed=1)


def test15_tuple_movable_in_value(clean):
    s = t.Movable()
    t.tuple_movable_in_value((s,))
    assert s.value == 5
    del s
    assert_stats(
        default_constructed=1,
        copy_constructed=1,
        destructed=2)


def test16_tuple_movable_in_lvalue_ref(clean):
    s = t.Movable()
    t.tuple_movable_in_lvalue_ref((s,))
    assert s.value == 5
    del s
    assert_stats(
        default_constructed=1,
        destructed=1)


def test17_tuple_movable_in_lvalue_ref_2(clean):
    s = t.Movable()
    t.tuple_movable_in_lvalue_ref_2((s,))
    assert s.value == 5
    del s
    assert_stats(
        default_constructed=1,
        copy_constructed=1,
        destructed=2)


def test18_tuple_movable_in_ptr(clean):
    s = t.Movable()
    t.tuple_movable_in_ptr((s,))
    assert s.value == 5
    del s
    assert_stats(
        default_constructed=1,
        destructed=1)


def test19_tuple_movable_in_rvalue_ref(clean):
    s = t.Movable()
    t.tuple_movable_in_rvalue_ref((s,))
    assert s.value == 0
    del s
    assert_stats(
        default_constructed=1,
        move_constructed=1,
        destructed=2)


def test20_tuple_movable_in_rvalue_ref_2(clean):
    s = t.Movable()
    t.tuple_movable_in_rvalue_ref_2((s,))
    assert s.value == 5
    del s
    assert_stats(
        default_constructed=1,
        copy_constructed=1,
        move_constructed=1,
        destructed=3)

# ------------------------------------------------------------------

def test21_tuple_pair_basic():
    assert t.empty_tuple(()) == ()
    assert t.swap_tuple((1, 2.5)) == (2.5, 1)
    assert t.swap_pair((1, 2.5)) == (2.5, 1)

# ------------------------------------------------------------------

def test22_vec_return_movable(clean):
    for i, x in enumerate(t.vec_return_movable()):
        assert x.value == i
    del x
    assert_stats(
        value_constructed=10,
        move_constructed=10,
        destructed=20
    )


def test23_vec_return_copyable(clean):
    for i, x in enumerate(t.vec_return_copyable()):
        assert x.value == i
    del x
    assert_stats(
        value_constructed=10,
        copy_constructed=20,
        destructed=30
    )


def test24_vec_movable_in_value(clean):
    t.vec_moveable_in_value([t.Movable(i) for i in range(10)])
    assert_stats(
        value_constructed=10,
        copy_constructed=10,
        move_constructed=10,
        destructed=30
    )


def test25_vec_movable_in_value(clean):
    t.vec_copyable_in_value([t.Copyable(i) for i in range(10)])
    assert_stats(
        value_constructed=10,
        copy_constructed=20,
        destructed=30
    )


def test26_vec_movable_in_lvalue_ref(clean):
    t.vec_moveable_in_lvalue_ref([t.Movable(i) for i in range(10)])
    assert_stats(
        value_constructed=10,
        move_constructed=10,
        destructed=20
    )


def test27_vec_movable_in_ptr_2(clean):
    t.vec_moveable_in_ptr_2([t.Movable(i) for i in range(10)])
    assert_stats(
        value_constructed=10,
        destructed=10
    )


def test28_vec_movable_in_rvalue_ref(clean):
    t.vec_moveable_in_rvalue_ref([t.Movable(i) for i in range(10)])
    assert_stats(
        value_constructed=10,
        move_constructed=10,
        destructed=20
    )


def test29_opaque_vector():
    f = t.float_vec()
    assert f.size() == 0
    assert isinstance(f, t.float_vec)
    f.push_back(1)
    assert f.size() == 1


def test30_std_function():
    assert t.return_empty_function() is None
    assert t.return_function()(3) == 8
    assert t.call_function(lambda x: 5 + x, 3) == 8

    with pytest.raises(TypeError) as excinfo:
        assert t.call_function(5, 3) == 8
    assert 'incompatible function arguments' in str(excinfo.value)

    with pytest.raises(TypeError) as excinfo:
        assert t.call_function(lambda x, y: x+y, 3) == 8
    assert 'missing 1 required positional argument' in str(excinfo.value)

    l = []
    def f():
        l.append(1)
    f2 = t.return_void_function(f)
    f2()
    assert l == [1]


def test31_std_function_roundtrip():
    def f():
        l.append(1)
    f2 = t.return_void_function(f)
    assert f2 is f


# cpyext reference cycles are not supported, see https://foss.heptapod.net/pypy/pypy/-/issues/3849
@skip_on_pypy
def test32_std_function_gc():
    class A(t.FuncWrapper):
        pass
    assert t.FuncWrapper.alive == 0
    a = A()
    assert t.FuncWrapper.alive == 1
    del a
    collect()
    assert t.FuncWrapper.alive == 0

    # A class -> function -> class cyclic reference
    class B(t.FuncWrapper):
        def __init__(self):
            super().__init__()
            def f():
                print(self.f)

            self.f = f
    assert t.FuncWrapper.alive == 0
    b = B()
    assert t.FuncWrapper.alive == 1
    del b
    collect()
    assert t.FuncWrapper.alive == 0


def test33_vec_type_check():
    with pytest.raises(TypeError) as excinfo:
        t.vec_moveable_in_value(0)


def test34_list():
    assert t.identity_list([]) == []
    assert t.identity_list([1, 2, 3]) == [1, 2, 3]
    assert t.identity_list(()) == []
    assert t.identity_list((1, 2, 3)) == [1, 2, 3]

def test35_string_and_string_view():
    assert t.identity_string("") == ""
    assert t.identity_string("orange") == "orange"
    assert t.identity_string("橘子") == "橘子"
    assert t.identity_string("ส้ม") == "ส้ม"
    assert t.identity_string("البرتقالي") == "البرتقالي"
    assert t.identity_string("🍊") == "🍊"

    assert t.identity_string_view("") == ""
    assert t.identity_string_view("orange") == "orange"
    assert t.identity_string_view("橘子") == "橘子"
    assert t.identity_string_view("ส้ม") == "ส้ม"
    assert t.identity_string_view("البرتقالي") == "البرتقالي"
    assert t.identity_string_view("🍊") == "🍊"


def test36_std_optional_copyable(clean):
    t.optional_copyable(t.Copyable())
    assert t.optional_copyable.__doc__ == (
        "optional_copyable(x: Optional[test_stl_ext.Copyable]) -> None"
    )
    assert_stats(
        default_constructed=1,
        copy_constructed=1,
        destructed=2
    )

def test37_std_optional_copyable_ptr(clean):
    t.optional_copyable_ptr(t.Copyable())
    assert t.optional_copyable_ptr.__doc__ == (
        "optional_copyable_ptr(x: Optional[test_stl_ext.Copyable]) -> None"
    )
    assert_stats(
        default_constructed=1,
        destructed=1
    )


def test38_std_optional_none():
    t.optional_none(None)


def test39_std_optional_ret_opt_movable(clean):
    assert t.optional_ret_opt_movable().value == 5
    assert t.optional_ret_opt_movable.__doc__ == (
        "optional_ret_opt_movable() -> Optional[test_stl_ext.Movable]"
    )
    assert_stats(
        default_constructed=1,
        move_constructed=2,
        destructed=3
    )

def test40_std_optional_ret_opt_movable_ptr(clean):
    assert t.optional_ret_opt_movable_ptr().value == 5
    assert_stats(
        default_constructed=1,
        destructed=1
    )


def test41_std_optional_ret_opt_none():
    assert t.optional_ret_opt_none() is None


def test42_std_optional_unbound_type():
    assert t.optional_unbound_type(3) == 3
    assert t.optional_unbound_type(None) is None
    assert t.optional_unbound_type() is None
    assert t.optional_unbound_type.__doc__ == (
        "optional_unbound_type(x: Optional[int] = None) -> Optional[int]"
    )


def test43_std_variant_copyable(clean):
    t.variant_copyable(t.Copyable())
    t.variant_copyable(5)
    assert t.variant_copyable.__doc__ == (
        "variant_copyable(arg: Union[test_stl_ext.Copyable, int], /) -> None"
    )
    assert_stats(
        default_constructed=3,
        copy_assigned=1,
        destructed=3
    )


def test44_std_variant_copyable_none(clean):
    t.variant_copyable_none(t.Copyable())
    t.variant_copyable_none(5)
    t.variant_copyable_none(None)
    assert t.variant_copyable_none.__doc__ == (
        "variant_copyable_none(x: Optional[Union[test_stl_ext.Copyable, int]]) -> None"
    )
    assert_stats(
        default_constructed=1,
        copy_constructed=1,
        destructed=2
    )


def test45_std_variant_copyable_ptr(clean):
    t.variant_copyable_ptr(t.Copyable())
    t.variant_copyable_ptr(5)
    assert t.variant_copyable_ptr.__doc__ == (
        "variant_copyable_ptr(arg: Union[test_stl_ext.Copyable, int], /) -> None"
    )
    assert_stats(
        default_constructed=1,
        destructed=1
    )


def test46_std_variant_copyable_ptr_none(clean):
    t.variant_copyable_ptr_none(t.Copyable())
    t.variant_copyable_ptr_none(5)
    t.variant_copyable_ptr_none(None)
    assert t.variant_copyable_ptr_none.__doc__ == (
        "variant_copyable_ptr_none(x: Optional[Union[test_stl_ext.Copyable, int]]) -> None"
    )
    assert_stats(
        default_constructed=1,
        destructed=1
    )

def test47_std_variant_ret_var_copyable():
    assert t.variant_ret_var_copyable().value == 5
    assert t.variant_ret_var_copyable.__doc__ == (
        "variant_ret_var_copyable() -> Union[test_stl_ext.Copyable, int]"
    )


def test48_std_variant_ret_var_none():
    assert t.variant_ret_var_none() is None
    assert t.variant_ret_var_none.__doc__ == (
        "variant_ret_var_none() -> Union[None, test_stl_ext.Copyable, int]"
    )


def test49_std_variant_unbound_type():
    assert t.variant_unbound_type() is None
    assert t.variant_unbound_type(None) is None
    assert t.variant_unbound_type([5]) == [5]
    assert t.variant_unbound_type((1,2,3)) == (1,2,3)
    assert t.variant_unbound_type(5) == 5
    assert t.variant_unbound_type.__doc__ == (
        "variant_unbound_type(x: Optional[Union[list, tuple, int]] = None)"
        " -> Union[None, list, tuple, int]"
    )

def test50_map_return_movable_value():
    for i, (k, v) in enumerate(sorted(t.map_return_movable_value().items())):
        assert k == chr(ord("a") + i)
        assert v.value == i
    if sys.version_info < (3, 9):
        assert t.map_return_movable_value.__doc__ == (
            "map_return_movable_value() -> Dict[str, test_stl_ext.Movable]"
        )
    else:
        assert t.map_return_movable_value.__doc__ == (
            "map_return_movable_value() -> dict[str, test_stl_ext.Movable]"
        )


def test51_map_return_copyable_value():
    for i, (k, v) in enumerate(sorted(t.map_return_copyable_value().items())):
        assert k == chr(ord("a") + i)
        assert v.value == i
    if sys.version_info < (3, 9):
        assert t.map_return_copyable_value.__doc__ == (
            "map_return_copyable_value() -> Dict[str, test_stl_ext.Copyable]"
        )
    else:
        assert t.map_return_copyable_value.__doc__ == (
            "map_return_copyable_value() -> dict[str, test_stl_ext.Copyable]"
        )

def test52_map_movable_in_value():
    t.map_movable_in_value(dict([(chr(ord("a") + i), t.Movable(i)) for i in range(10)]))
    if sys.version_info < (3, 9):
        assert t.map_movable_in_value.__doc__ == (
            "map_movable_in_value(x: Dict[str, test_stl_ext.Movable]) -> None"
        )
    else:
        assert t.map_movable_in_value.__doc__ == (
            "map_movable_in_value(x: dict[str, test_stl_ext.Movable]) -> None"
        )


def test53_map_copyable_in_value():
    t.map_copyable_in_value(dict([(chr(ord("a") + i), t.Copyable(i)) for i in range(10)]))
    if sys.version_info < (3, 9):
        assert t.map_copyable_in_value.__doc__ == (
            "map_copyable_in_value(x: Dict[str, test_stl_ext.Copyable]) -> None"
        )
    else:
        assert t.map_copyable_in_value.__doc__ == (
            "map_copyable_in_value(x: dict[str, test_stl_ext.Copyable]) -> None"
        )


def test54_map_movable_in_lvalue_ref():
    t.map_movable_in_lvalue_ref(dict([(chr(ord("a") + i), t.Movable(i)) for i in range(10)]))
    if sys.version_info < (3, 9):
        assert t.map_movable_in_lvalue_ref.__doc__ == (
            "map_movable_in_lvalue_ref(x: Dict[str, test_stl_ext.Movable]) -> None"
        )
    else:
        assert t.map_movable_in_lvalue_ref.__doc__ == (
            "map_movable_in_lvalue_ref(x: dict[str, test_stl_ext.Movable]) -> None"
        )


def test55_map_movable_in_rvalue_ref():
    t.map_movable_in_rvalue_ref(dict([(chr(ord("a") + i), t.Movable(i)) for i in range(10)]))
    if sys.version_info < (3, 9):
        assert t.map_movable_in_rvalue_ref.__doc__ == (
            "map_movable_in_rvalue_ref(x: Dict[str, test_stl_ext.Movable]) -> None"
        )
    else:
        assert t.map_movable_in_rvalue_ref.__doc__ == (
            "map_movable_in_rvalue_ref(x: dict[str, test_stl_ext.Movable]) -> None"
        )


def test56_map_movable_in_ptr():
    t.map_movable_in_ptr(dict([(chr(ord("a") + i), t.Movable(i)) for i in range(10)]))
    if sys.version_info < (3, 9):
        assert t.map_movable_in_ptr.__doc__ == (
            "map_movable_in_ptr(x: Dict[str, test_stl_ext.Movable]) -> None"
        )
    else:
        assert t.map_movable_in_ptr.__doc__ == (
            "map_movable_in_ptr(x: dict[str, test_stl_ext.Movable]) -> None"
        )


def test57_map_return_readonly_value():
    for i, (k, v) in enumerate(sorted(t.map_return_readonly_value().map.items())):
        assert k == chr(ord("a") + i)
        assert v == i
    assert t.map_return_readonly_value.__doc__ == (
        "map_return_readonly_value() -> test_stl_ext.StructWithReadonlyMap"
    )


def test58_array():
    o = t.array_out()
    assert isinstance(o, list) and o == [1, 2, 3]
    assert t.array_in([1, 2, 3]) == 6
    assert t.array_in((1, 2, 3)) == 6
    with pytest.raises(TypeError) as excinfo:
        assert t.array_in((1, 2, 3, 4)) == 6
    assert 'incompatible function arguments' in str(excinfo.value)

def test59_map_movable_in_failure():
    with pytest.raises(TypeError) as excinfo:
        t.map_copyable_in_value({1:2})
    assert 'incompatible function arguments' in str(excinfo.value)


def test60_set_return_value():
    for i, k in enumerate(sorted(t.set_return_value())):
        assert k == chr(ord("a") + i)
    for i, k in enumerate(sorted(t.unordered_set_return_value())):
        assert k == chr(ord("a") + i)
    if sys.version_info < (3, 9):
        assert t.set_return_value.__doc__ == (
            "set_return_value() -> Set[str]")
        assert t.unordered_set_return_value.__doc__ == (
            "unordered_set_return_value() -> Set[str]")
    else:
        assert t.set_return_value.__doc__ == (
            "set_return_value() -> set[str]")
        assert t.unordered_set_return_value.__doc__ == (
            "unordered_set_return_value() -> set[str]")

def test61_set_in_value():
    t.set_in_value(set([chr(ord("a") + i) for i in range(10)]))
    t.unordered_set_in_value(set([chr(ord("a") + i) for i in range(10)]))
    if sys.version_info < (3, 9):
        assert t.set_in_value.__doc__ == (
            "set_in_value(x: Set[str]) -> None")
        assert t.unordered_set_in_value.__doc__ == (
            "unordered_set_in_value(x: Set[str]) -> None")
    else:
        assert t.set_in_value.__doc__ == (
            "set_in_value(x: set[str]) -> None")
        assert t.unordered_set_in_value.__doc__ == (
            "unordered_set_in_value(x: set[str]) -> None")


def test62_set_in_lvalue_ref():
    t.set_in_lvalue_ref(set([chr(ord("a") + i) for i in range(10)]))
    if sys.version_info < (3, 9):
        assert t.set_in_lvalue_ref.__doc__ == (
            "set_in_lvalue_ref(x: Set[str]) -> None"
        )
    else:
        assert t.set_in_lvalue_ref.__doc__ == (
            "set_in_lvalue_ref(x: set[str]) -> None"
        )

def test63_set_in_rvalue_ref():
    t.set_in_rvalue_ref(set([chr(ord("a") + i) for i in range(10)]))
    if sys.version_info < (3, 9):
        assert t.set_in_rvalue_ref.__doc__ == (
            "set_in_rvalue_ref(x: Set[str]) -> None"
        )
    else:
        assert t.set_in_rvalue_ref.__doc__ == (
            "set_in_rvalue_ref(x: set[str]) -> None"
        )


def test64_set_in_failure():
    with pytest.raises(TypeError) as excinfo:
        t.set_in_value(set([i for i in range(10)]))
    assert 'incompatible function arguments' in str(excinfo.value)


def test65_class_with_movable_field(clean):
    cwmf = t.ClassWithMovableField()
    m1 = t.Movable(1)
    m2 = t.Movable(2)

    assert_stats(
        value_constructed=2
    )

    cwmf.movable = [ m1, m2 ]

    assert_stats(
        value_constructed=2,
        move_constructed=2
    )

    del m1, m2
    collect()

    assert_stats(
        value_constructed=2,
        move_constructed=2,
        destructed=2
    )

    del cwmf
    collect()

    assert_stats(
        value_constructed=2,
        move_constructed=2,
        destructed=4
    )

def test66_replace_extension():
    from pathlib import Path

    if not hasattr(t, 'replace_extension'):
        pytest.skip('filesystem tests not compiled')

    filename = Path("test.txt")
    assert t.replace_extension(filename, ".obj") == filename.with_suffix(".obj")

    filename = Path("🍊.html")
    assert t.replace_extension(filename, ".svg") == filename.with_suffix(".svg")

    class PseudoStrPath:
        def __fspath__(self):
            return "foo/bar"

    class PseudoBytesPath:
        def __fspath__(self):
            return b"foo/bar"

    assert t.parent_path(Path("foo/bar")) == Path("foo")
    assert t.parent_path("foo/bar") == Path("foo")
    assert t.parent_path(b"foo/bar") == Path("foo")
    assert t.parent_path(PseudoStrPath()) == Path("foo")
    assert t.parent_path(PseudoBytesPath()) == Path("foo")




import sys, re, os

# g_lib_name = "/opt/conda/lib/python3.10/site-packages/torch/lib/libtorch_cuda.so"
def addr2line(addr):
    s = f"addr2line -e {g_lib_name} -f {addr}"
    with os.popen(s) as pf:
        symbol = pf.readlines()
        return [s.strip() for s in symbol]

def demangler(symbol):
    if len(symbol) == 0:
        return ''
    end = symbol.find("+0x")
    if end != -1:
        symbol = symbol[:end]
    else:
        return ''
    s = f"c++filt {symbol}"
    with os.popen(s) as pf:
        symbol = pf.readlines()[0]
        return symbol.strip()

def parse_and_filter(file):
    new_lines = []
    with open(file, "rt") as f:
        lines = f.readlines()
        first_line = lines[0]

        base_addr = first_line[len("ignore:[base address]:"):]
        base_addr_int = int(base_addr, 16)
        for line in lines[1:]:
            if line.find("ignore") != -1:
                lines.append(line)
                continue
            if line.find("libtorch_cuda.so") == -1:
                lines.append(line)
                continue
            def repl(g):
                file_addr = hex(int(g.group(3), 16) - base_addr_int)
                # info = addr2line(g.group(1)[1:])[0]
                info = addr2line(file_addr)
                symbol = info[0]
                sorce_file = info[1]
                result = g.group(0).replace(g.group(1), sorce_file)
                if g.group(2)[0] == "+":
                    result = result.replace(g.group(2), symbol)
                return result
            line = re.sub(r"(.*libtorch_cuda.so)\((.*)\) \[(.*?)\]", repl, line)
            new_lines.append(line)
    with open(file, "wt") as f:
        for line in new_lines:
            f.write(line)

def main():
    g_lib_name = sys.argv[1]
    file = sys.argv[2]
    file = sys.argv[1]
    parse_and_filter(file)
    
    with open(file, "rt") as f:
        lines = []
        for line in f.readlines():
            if line.find("ignore") != -1:
                lines.append(line)
                continue
            def repl(g):
                return g.group(0).replace(g.group(1), demangler(g.group(1)))
            line = re.sub(r".*?\(([\s|\S]*?)[\)|\n]", repl, line, re.M)
            lines.append(line)
    with open(file, "wt") as f:
        for line in lines:
            f.write(line)



if __name__ == "__main__":
    main()
